#include "gebs/gebs.h"
#include "mongoose/mongoose.h"
#include "cJSON/cJSON.h"
#if MY_DEBUG
#include "md5-c/md5.h"
#endif

#include "routes.h"
#include "baked.h"
#include "clonestr.h"
#include "commit.h"
#include "timer.h"

#define INTERNAL_SERVER_ERROR_MSG "Internal server error ;(. Try again later..."

void make_internal_server_error(Route_Result *result)
{
    result->status_code = 500;
    result->type = ROUTE_RESULT_DYNAMIC;
    list_append(&result->headers, "Content-Type: text/plain");
    sb_append_nstr(&result->body, INTERNAL_SERVER_ERROR_MSG);
    sb_finish(&result->body);
}

void make_application_json(Route_Result *result, int code, cJSON *root)
{
    char *root_text = cJSON_PrintUnformatted(root);
    defer { free(root_text); }

    result->status_code = code;
    result->type = ROUTE_RESULT_DYNAMIC;
    list_append(&result->headers, clonestr("Content-Type: application/json"));
    sb_append_nstr(&result->body, root_text);
    sb_finish(&result->body);
}

void make_binary(Route_Result *result, char *content_type, int code, char *data, size_t size)
{
    char type[100];
    snprintf(type, sizeof(type), "Content-Type: %s", content_type);
    result->status_code = code;
    result->type = ROUTE_RESULT_DYNAMIC;
    list_append(&result->headers, clonestr(type));
    for (size_t i = 0; i < size; i++) {
        sb_append_char(&result->body, data[i]);
    }
    sb_finish(&result->body);
}

void make_text(Route_Result *result, char *subtype, int code, char *in)
{
    char type[100];
    snprintf(type, sizeof(type), "Content-Type: text/%s", subtype);

    result->status_code = code;
    result->type = ROUTE_RESULT_DYNAMIC;
    list_append(&result->headers, clonestr(type));
    sb_append_nstr(&result->body, in);
    sb_finish(&result->body);
}

bool gpp_run(char *path, NString_List *env, String_Builder *out)
{
    Cmd cmd = {0};
    defer { cmd_free(&cmd); }

    char gpp1[PATH_MAX];
    if (!get_baked_resource_path("gpp1", gpp1, sizeof(gpp1))) {
        return false;
    }

    cmd_append(&cmd, gpp1);
    cmd_append(&cmd, "-H");
    cmd_append(&cmd, "-x");
    cmd_append(&cmd, "--nostdinc");
    cmd_append(&cmd, path);

    for (size_t i = 0; i < env->count; i++) {
        cmd_append(&cmd, env->items[i]);
    }

    return cmd_run_collect(&cmd, out) == 0;
}

#if MY_DEBUG
ROUTE_HANDLER(build_id)
{
    cJSON *root = cJSON_CreateObject();
    defer { cJSON_Delete(root); }

    char *time = __TIME__;
    uchar md5_buf[16];
    md5String(time, md5_buf);
    String_Builder sb = {0};
    defer { sb_free(&sb); }
    for (size_t i = 0; i < 16; i++) {
        sb_append_nstr(&sb, fmt("%02x", md5_buf[i]));
    }
    sb_finish(&sb);

    cJSON_AddItemToObject(root, "build_id", cJSON_CreateString(sb.items));

    make_application_json(result, 200, root);
}
#endif

ROUTE_HANDLER(page_not_found)
{
    NString_List env = {0};
    defer { list_free(&env); }

    list_append(&env, fmt("-DURL=%.*s", msg->uri.len, msg->uri.buf));

    String_Builder out = {0};
    defer { sb_free(&out); }
    
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("page-missing.html", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }
    
    list_append(&env, fmt("-DCOMMIT=%s", COMMIT_STRING));
    char timer[100];
    get_timer_string(timer, sizeof(timer));
    list_append(&env, fmt("-DRUNNING_SINCE=%s", timer));

    #if MY_DEBUG
        list_append(&env, "-DHOTRELOAD");
    #endif

    bool ok = gpp_run(path, &env, &out);
    sb_finish(&out);

    if (!ok) {
        make_internal_server_error(result);
    } else {
        make_text(result, "html", 200, out.items);
    }
}

ROUTE_HANDLER(simple_css)
{
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("simple.css", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    String_Builder sb = {0};
    defer { sb_free(&sb); }
    if (!sb_read_file(&sb, path)) {
        make_internal_server_error(result);
        return;
    }
    sb_finish(&sb);

    make_text(result, "css", 200, sb.items);
}

ROUTE_HANDLER(me_jpg)
{
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("me.jpg", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    String_Builder sb = {0};
    defer { sb_free(&sb); }
    if (!sb_read_file(&sb, path)) {
        make_internal_server_error(result);
        return;
    }

    make_binary(result, "image/jpeg", 200, sb.items, sb.count);
}

ROUTE_HANDLER(favicon)
{
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("favicon.ico", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    String_Builder sb = {0};
    defer { sb_free(&sb); }
    if (!sb_read_file(&sb, path)) {
        make_internal_server_error(result);
        return;
    }

    make_binary(result, "image/x-icon", 200, sb.items, sb.count);
}

#if MY_DEBUG
ROUTE_HANDLER(hotreload_js)
{
    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("hotreload.js", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }

    String_Builder sb = {0};
    defer { sb_free(&sb); }
    if (!sb_read_file(&sb, path)) {
        make_internal_server_error(result);
        return;
    }
    sb_finish(&sb);

    make_text(result, "javascript", 200, sb.items);
}
#endif

ROUTE_HANDLER(home)
{
    NString_List env = {0};
    defer { list_free(&env); }

    String_Builder out = {0};
    defer { sb_free(&out); }

    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("home.html", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }
    
    list_append(&env, fmt("-DCOMMIT=%s", COMMIT_STRING));
    char timer[100];
    get_timer_string(timer, sizeof(timer));
    list_append(&env, fmt("-DRUNNING_SINCE=%s", timer));

    #if MY_DEBUG
        list_append(&env, "-DHOTRELOAD");
    #endif

    bool ok = gpp_run(path, &env, &out);
    sb_finish(&out);

    if (!ok) {
        make_internal_server_error(result);
    } else {
        make_text(result, "html", 200, out.items);
    }
}

ROUTE_HANDLER(generic_blog)
{
    NString_List env = {0};
    defer { list_free(&env); }

    String_Builder out = {0};
    defer { sb_free(&out); }

    Baked_Resource *resource = (Baked_Resource *)context_data;

    char md_path[PATH_MAX] = {0};
    if (!get_baked_resource_path(resource->key, md_path, sizeof(md_path))) {
        make_internal_server_error(result);
        return;
    }
    
    list_append(&env, fmt("-DCOMMIT=%s", COMMIT_STRING));
    char timer[100];
    get_timer_string(timer, sizeof(timer));
    list_append(&env, fmt("-DRUNNING_SINCE=%s", timer));
    
    #if MY_DEBUG
        list_append(&env, "-DHOTRELOAD");
    #endif

    String_Builder md = {0};
    defer { sb_free(&md); }
    if (!sb_read_file(&md, md_path)) {
        make_internal_server_error(result);
        return;
    }

    String_Builder md_prepared = {0};
    defer { sb_free(&md_prepared); }
    for (size_t i = 0; i < md.count; i++) {
        char c = md.items[i];
        switch (c) {
            case '\"':  sb_append_nstr(&md_prepared, "\\\"");    break;
            case '\'':  sb_append_nstr(&md_prepared, "\\\'");    break;
            case '\\':  sb_append_nstr(&md_prepared, "\\\\");    break;
            case '\a':  sb_append_nstr(&md_prepared, "\\a");     break;
            case '\b':  sb_append_nstr(&md_prepared, "\\b");     break;
            case '\r':  sb_append_nstr(&md_prepared, "\\r");     break;
            case '\n':  sb_append_nstr(&md_prepared, "\\n");     break;
            case '\t':  sb_append_nstr(&md_prepared, "\\t");     break;
            default:
                if (iscntrl(md.items[i])) {
                    sb_append_nstr(&md_prepared, fmt("\\%03o", c));
                } else {
                    sb_append_char(&md_prepared, c);
                }
                break;
        }
    }
    sb_finish(&md_prepared);

    list_append(&env, fmt("-DBLOG_POST_TITLE=%s", resource->key));
    list_append(&env, fmt("-DBLOG_POST_MARKDOWN=%s", md_prepared.items));

    char tmpl_file[PATH_MAX] = {0};
    if (!get_baked_resource_path("template-blog.html", tmpl_file, sizeof(tmpl_file))) {
        make_internal_server_error(result);
        return;
    }

    bool ok = gpp_run(tmpl_file, &env, &out);
    sb_finish(&out);

    if (!ok) {
        make_internal_server_error(result);
    } else {
        make_text(result, "html", 200, out.items);
    }
}

void collect_blogs(Baked_Resource *resource, void *udata)
{
    struct { Arena *tmp; String_Builder *sb; } *ud = udata;
    if ((strlen(resource->key) >= strlen("blog-"))
     && strncmp(resource->key, "blog-", strlen("blog-")) == 0) {
        sb_append_nstr_alloc(ud->tmp, ud->sb, fmt("<li><a href=\"/%s\">%s</a></li>", resource->key, resource->key));
    }
}

ROUTE_HANDLER(blog)
{
    NString_List env = {0};
    defer { list_free(&env); }

    String_Builder out = {0};
    defer { sb_free(&out); }

    Arena tmp = arena_get();
    defer { arena_destroy(&tmp); }

    String_Builder sb = {0};
    struct { Arena *tmp; String_Builder *sb; } collect_blogs_data = {
        .tmp = &tmp,
        .sb = &sb,
    };
    baked_resource_each(&collect_blogs, &collect_blogs_data);
    sb_finish_alloc(&tmp, &sb);
    list_append(&env, fmt("-DBLOG_POSTS=%s", sb.items));

    char path[PATH_MAX] = {0};
    if (!get_baked_resource_path("blog.html", path, sizeof(path))) {
        make_internal_server_error(result);
        return;
    }
    
    list_append(&env, fmt("-DCOMMIT=%s", COMMIT_STRING));
    char timer[100];
    get_timer_string(timer, sizeof(timer));
    list_append(&env, fmt("-DRUNNING_SINCE=%s", timer));

    #if MY_DEBUG
        list_append(&env, "-DHOTRELOAD");
    #endif

    bool ok = gpp_run(path, &env, &out);
    sb_finish(&out);

    if (!ok) {
        make_internal_server_error(result);
    } else {
        make_text(result, "html", 200, out.items);
    }
}

